@@ -26,7 +26,7 @@
 #include <math.h>
 #include <errno.h>
 
-#include <libtracker-common/tracker-common.h>
+#include <libtracker-common/tracker-locale.h>
 
 #include <libtracker-sparql/tracker-sparql.h>
 
@@ -37,6 +37,7 @@
 #include "tracker-collation.h"
 
 #include "tracker-db-interface-sqlite.h"
+#include "tracker-db-manager.h"
 
 #define UNKNOWN_STATUS 0.5
 
@@ -57,7 +58,13 @@ struct TrackerDBInterface {
 
 	GSList *function_data;
 
+	/* Collation and locale change */
 	gpointer collator;
+	gpointer locale_notification_id;
+	gint collator_reset_requested;
+
+	/* Number of active cursors */
+	gint n_active_cursors;
 
 	guint ro : 1;
 #if HAVE_TRACKER_FTS
@@ -65,10 +72,6 @@ struct TrackerDBInterface {
 #endif
 	GCancellable *cancellable;
 
-	/* async operation pending */
-	gboolean pending;
-	GAsyncReadyCallback outstanding_callback;
-
 	TrackerDBStatementLru select_stmt_lru;
 	TrackerDBStatementLru update_stmt_lru;
 
@@ -90,6 +93,8 @@ struct TrackerDBCursor {
 	gint n_types;
 	gchar **variable_names;
 	gint n_variable_names;
+
+	gboolean threadsafe;
 };
 
 struct TrackerDBCursorClass {
@@ -117,7 +122,8 @@ static TrackerDBCursor    *tracker_db_cursor_sqlite_new      (sqlite3_stmt
                                                               TrackerPropertyType  *types,
                                                               gint                  n_types,
                                                               const gchar         **variable_names,
-                                                              gint                  n_variable_names);
+                                                              gint                  n_variable_names,
+                                                              gboolean              threadsafe);
 static gboolean            tracker_db_cursor_get_boolean     (TrackerSparqlCursor  *cursor,
                                                               guint                 column);
 static gboolean            db_cursor_iter_next               (TrackerDBCursor      *cursor,
@@ -526,6 +532,17 @@ check_interrupt (void *user_data)
 	return g_cancellable_is_cancelled (db_interface->cancellable) ? 1 : 0;
 }
 
+static void
+tracker_locale_notify_cb (TrackerLocaleID id,
+                          gpointer        user_data)
+{
+	TrackerDBInterface *db_interface = user_data;
+
+	/* Request a collator reset. Use thread-safe methods as this function will get
+	 * called from the main thread */
+	g_atomic_int_compare_and_exchange (&(db_interface->collator_reset_requested), FALSE, TRUE);
+}
+
 static void
 open_database (TrackerDBInterface *db_interface)
 {
@@ -546,18 +563,12 @@ open_database (TrackerDBInterface *db_interface)
 	}
 
 	/* Set our unicode collation function */
-	if (!db_interface->collator) {
-		db_interface->collator = tracker_collation_init ();
-		if (sqlite3_create_collation (db_interface->db,
-		                              TRACKER_COLLATION_NAME,
-		                              SQLITE_UTF8,
-		                              db_interface->collator,
-		                              tracker_collation_utf8) != SQLITE_OK)
-		{
-			g_critical ("Couldn't set collation function: %s",
-			            sqlite3_errmsg (db_interface->db));
-		}
-	}
+	tracker_db_interface_sqlite_reset_collator (db_interface);
+	/* And register for updates on locale changes */
+	db_interface->locale_notification_id = tracker_locale_notify_add (TRACKER_LOCALE_COLLATE,
+	                                                                  tracker_locale_notify_cb,
+	                                                                  db_interface,
+	                                                                  NULL);
 
 	sqlite3_progress_handler (db_interface->db, 100,
 	                          check_interrupt, db_interface);
@@ -603,8 +614,8 @@ tracker_db_interface_sqlite_constructor (GType                  type,
 	TrackerDBInterface *db_iface;
 
 	object = (* G_OBJECT_CLASS (tracker_db_interface_parent_class)->constructor) (type,
-		               n_construct_properties,
-		               construct_params);
+		        n_construct_properties,
+		        construct_params);
 	db_iface = TRACKER_DB_INTERFACE (object);
 
 	open_database (db_iface);
@@ -720,6 +731,28 @@ tracker_db_interface_sqlite_fts_update_rollback (TrackerDBInterface *db_interfac
 }
 #endif
 
+void
+tracker_db_interface_sqlite_reset_collator (TrackerDBInterface *db_interface)
+{
+	g_debug ("Resetting collator in db interface %p", db_interface);
+	if (db_interface->collator)
+		tracker_collation_shutdown (db_interface->collator);
+
+	db_interface->collator = tracker_collation_init ();
+	/* This will overwrite any other collation set before, if any */
+	if (sqlite3_create_collation (db_interface->db,
+	                              TRACKER_COLLATION_NAME,
+	                              SQLITE_UTF8,
+	                              db_interface->collator,
+	                              tracker_collation_utf8) != SQLITE_OK)
+	{
+		g_critical ("Couldn't set collation function: %s",
+		            sqlite3_errmsg (db_interface->db));
+	}
+}
+
+
+
 static void
 tracker_db_interface_sqlite_finalize (GObject *object)
 {
@@ -734,6 +767,7 @@ tracker_db_interface_sqlite_finalize (GObject *object)
 	g_free (db_interface->filename);
 	g_free (db_interface->busy_status);
 
+	tracker_locale_notify_remove (db_interface->locale_notification_id);
 	tracker_collation_shutdown (db_interface->collator);
 
 	G_OBJECT_CLASS (tracker_db_interface_parent_class)->finalize (object);
@@ -782,65 +816,6 @@ tracker_db_interface_init (TrackerDBInterface *db_interface)
 	prepare_database (db_interface);
 }
 
-static TrackerDBResultSet *
-ensure_result_set_state (TrackerDBResultSet *result_set)
-{
-	if (!result_set)
-		return NULL;
-
-	if (tracker_db_result_set_get_n_rows (result_set) == 0) {
-		g_object_unref (result_set);
-		return NULL;
-	}
-
-	/* ensure that it's at the first item */
-	tracker_db_result_set_rewind (result_set);
-
-	return result_set;
-}
-
-static void
-add_row (TrackerDBResultSet *result_set,
-         sqlite3_stmt       *stmt)
-{
-	gint cols, i;
-
-	cols = sqlite3_column_count (stmt);
-	_tracker_db_result_set_append (result_set);
-
-	for (i = 0; i < cols; i++) {
-		GValue value = { 0, };
-		gint col_type;
-
-		col_type = sqlite3_column_type (stmt, i);
-
-		switch (col_type) {
-		case SQLITE_TEXT:
-			g_value_init (&value, G_TYPE_STRING);
-			g_value_set_string (&value, (gchar *) sqlite3_column_text (stmt, i));
-			break;
-		case SQLITE_INTEGER:
-			g_value_init (&value, G_TYPE_INT64);
-			g_value_set_int64 (&value, sqlite3_column_int64 (stmt, i));
-			break;
-		case SQLITE_FLOAT:
-			g_value_init (&value, G_TYPE_DOUBLE);
-			g_value_set_double (&value, sqlite3_column_double (stmt, i));
-			break;
-		case SQLITE_NULL:
-			/* just ignore NULLs */
-			break;
-		default:
-			g_critical ("Unknown sqlite3 database column type:%d", col_type);
-		}
-
-		if (G_VALUE_TYPE (&value) != G_TYPE_INVALID) {
-			_tracker_db_result_set_set_value (result_set, i, &value);
-			g_value_unset (&value);
-		}
-	}
-}
-
 void
 tracker_db_interface_set_max_stmt_cache_size (TrackerDBInterface         *db_interface,
                                               TrackerDBStatementCacheType cache_type,
@@ -895,18 +870,17 @@ tracker_db_interface_create_statement (TrackerDBInterface           *db_interfac
 	gchar *full_query;
 
 	g_return_val_if_fail (TRACKER_IS_DB_INTERFACE (db_interface), NULL);
-	g_return_val_if_fail (!db_interface->pending, NULL);
 
 	va_start (args, query);
 	full_query = g_strdup_vprintf (query, args);
 	va_end (args);
 
-	/* There are three kinds of queries: 
-	  * a) Cached queries: SELECT and UPDATE ones (cache_type)
-	  * b) Non-Cached queries: NONE ones (cache_type)
-	  * c) Forced Non-Cached: in case of a stmt being already in use, we can't
-	  *    reuse it (you can't use two different loops on a sqlite3_stmt, of
-	  *    course). This happens with recursive uses of a cursor, for example */
+	/* There are three kinds of queries:
+	 * a) Cached queries: SELECT and UPDATE ones (cache_type)
+	 * b) Non-Cached queries: NONE ones (cache_type)
+	 * c) Forced Non-Cached: in case of a stmt being already in use, we can't
+	 *    reuse it (you can't use two different loops on a sqlite3_stmt, of
+	 *    course). This happens with recursive uses of a cursor, for example */
 
 	if (cache_type != TRACKER_DB_STATEMENT_CACHE_TYPE_NONE) {
 		stmt = g_hash_table_lookup (db_interface->dynamic_statements, full_query);
@@ -1050,18 +1024,24 @@ tracker_db_interface_create_statement (TrackerDBInterface           *db_interfac
 	return (cache_type != TRACKER_DB_STATEMENT_CACHE_TYPE_NONE) ? g_object_ref (stmt) : stmt;
 }
 
-static TrackerDBResultSet *
-create_result_set_from_stmt (TrackerDBInterface  *interface,
-                             sqlite3_stmt        *stmt,
-                             GCancellable        *cancellable,
-                             GError             **error)
+static void
+execute_stmt (TrackerDBInterface  *interface,
+              sqlite3_stmt        *stmt,
+              GCancellable        *cancellable,
+              GError             **error)
 {
-	TrackerDBResultSet *result_set = NULL;
 	gint columns, result;
 
 	columns = sqlite3_column_count (stmt);
 	result = SQLITE_OK;
 
+	/* Statement is going to start, check if we got a request to reset the
+	 * collator, and if so, do it. */
+	if (g_atomic_int_exchange_and_add (&interface->n_active_cursors, 1) == 0 &&
+	    g_atomic_int_compare_and_exchange (&(interface->collator_reset_requested), TRUE, FALSE)) {
+		tracker_db_interface_sqlite_reset_collator (interface);
+	}
+
 	while (result == SQLITE_OK  ||
 	       result == SQLITE_ROW) {
 
@@ -1081,18 +1061,23 @@ create_result_set_from_stmt (TrackerDBInterface  *interface,
 			sqlite3_reset (stmt);
 			break;
 		case SQLITE_ROW:
-			if (G_UNLIKELY (!result_set)) {
-				result_set = _tracker_db_result_set_new (columns);
-			}
-
-			add_row (result_set, stmt);
 			break;
 		default:
 			break;
 		}
 	}
 
-	if (result != SQLITE_DONE) {
+
+	if (result == SQLITE_DONE) {
+		/* Statement finished, check if we got a request to reset the
+		 * collator, and if so, do it.
+		 */
+		if (g_atomic_int_dec_and_test (&interface->n_active_cursors) &&
+		    g_atomic_int_compare_and_exchange (&(interface->collator_reset_requested), TRUE, FALSE)) {
+			tracker_db_interface_sqlite_reset_collator (interface);
+		}
+	} else {
+		g_atomic_int_add (&interface->n_active_cursors, -1);
 
 		/* This is rather fatal */
 		if (sqlite3_errcode (interface->db) == SQLITE_IOERR ||
@@ -1115,7 +1100,7 @@ create_result_set_from_stmt (TrackerDBInterface  *interface,
 			         "Shutting down now.",
 			         interface->filename);
 
-			return NULL;
+			return;
 		}
 
 		if (!error) {
@@ -1136,26 +1121,16 @@ create_result_set_from_stmt (TrackerDBInterface  *interface,
 				             sqlite3_errmsg (interface->db));
 			}
 		}
-
-		/* If there was an error, result set may be invalid or incomplete */
-		if (result_set) {
-			g_object_unref (result_set);
-		}
-
-		return NULL;
 	}
-
-	return ensure_result_set_state (result_set);
 }
 
-TrackerDBResultSet *
+void
 tracker_db_interface_execute_vquery (TrackerDBInterface  *db_interface,
                                      GError             **error,
                                      const gchar         *query,
                                      va_list              args)
 {
 	gchar *full_query;
-	TrackerDBResultSet *result_set;
 	sqlite3_stmt *stmt;
 	int retval;
 
@@ -1171,7 +1146,7 @@ tracker_db_interface_execute_vquery (TrackerDBInterface  *db_interface,
 		             "%s",
 		             sqlite3_errmsg (db_interface->db));
 		g_free (full_query);
-		return NULL;
+		return;
 	} else if (stmt == NULL) {
 		g_set_error (error,
 		             TRACKER_DB_INTERFACE_ERROR,
@@ -1180,14 +1155,13 @@ tracker_db_interface_execute_vquery (TrackerDBInterface  *db_interface,
 		             full_query);
 
 		g_free (full_query);
-		return NULL;
+		return;
 	}
 
-	result_set = create_result_set_from_stmt (db_interface, stmt, NULL, error);
+	execute_stmt (db_interface, stmt, NULL, error);
 	sqlite3_finalize (stmt);
 
 	g_free (full_query);
-	return result_set;
 }
 
 TrackerDBInterface *
@@ -1261,10 +1235,19 @@ static void
 tracker_db_cursor_finalize (GObject *object)
 {
 	TrackerDBCursor *cursor;
+	TrackerDBInterface *iface;
 	int i;
 
 	cursor = TRACKER_DB_CURSOR (object);
 
+	/* As soon as we finalize the cursor, check if we need a collator reset
+	 * and notify the iface about the removed cursor */
+	iface = cursor->ref_stmt->db_interface;
+	if (g_atomic_int_dec_and_test (&iface->n_active_cursors) &&
+	    g_atomic_int_compare_and_exchange (&(iface->collator_reset_requested), TRUE, FALSE)) {
+		tracker_db_interface_sqlite_reset_collator (iface);
+	}
+
 	g_free (cursor->types);
 
 	for (i = 0; i < cursor->n_variable_names; i++) {
@@ -1272,13 +1255,9 @@ tracker_db_cursor_finalize (GObject *object)
 	}
 	g_free (cursor->variable_names);
 
-	if (cursor->ref_stmt) {
-		cursor->ref_stmt->stmt_is_sunk = FALSE;
-		tracker_db_statement_sqlite_reset (cursor->ref_stmt);
-		g_object_unref (cursor->ref_stmt);
-	} else {
-		sqlite3_finalize (cursor->stmt);
-	}
+	cursor->ref_stmt->stmt_is_sunk = FALSE;
+	tracker_db_statement_sqlite_reset (cursor->ref_stmt);
+	g_object_unref (cursor->ref_stmt);
 
 	G_OBJECT_CLASS (tracker_db_cursor_parent_class)->finalize (object);
 }
@@ -1302,24 +1281,6 @@ tracker_db_cursor_iter_next_thread (GSimpleAsyncResult *res,
 	}
 }
 
-static void
-async_ready_callback_wrapper (GObject      *source_object,
-                              GAsyncResult *res,
-                              gpointer      user_data)
-{
-	TrackerDBCursor *cursor = TRACKER_DB_CURSOR (source_object);
-	TrackerDBInterface *iface = cursor->ref_stmt->db_interface;
-
-	iface->pending = FALSE;
-	if (iface->outstanding_callback) {
-		GAsyncReadyCallback callback = iface->outstanding_callback;
-
-		iface->outstanding_callback = NULL;
-		(callback) (source_object, res, user_data);
-	}
-	g_object_unref (cursor);
-}
-
 static void
 tracker_db_cursor_iter_next_async (TrackerDBCursor     *cursor,
                                    GCancellable        *cancellable,
@@ -1328,12 +1289,9 @@ tracker_db_cursor_iter_next_async (TrackerDBCursor     *cursor,
 {
 	GSimpleAsyncResult *res;
 
-	g_return_if_fail (!cursor->ref_stmt->db_interface->pending);
-	cursor->ref_stmt->db_interface->pending = TRUE;
-	cursor->ref_stmt->db_interface->outstanding_callback = callback;
-
-	res = g_simple_async_result_new (g_object_ref (cursor), async_ready_callback_wrapper, user_data, tracker_db_cursor_iter_next_async);
+	res = g_simple_async_result_new (G_OBJECT (cursor), callback, user_data, tracker_db_cursor_iter_next_async);
 	g_simple_async_result_run_in_thread (res, tracker_db_cursor_iter_next_thread, 0, cancellable);
+	g_object_unref (res);
 }
 
 static gboolean
@@ -1375,21 +1333,29 @@ tracker_db_cursor_sqlite_new (sqlite3_stmt        *sqlite_stmt,
                               TrackerPropertyType *types,
                               gint                 n_types,
                               const gchar        **variable_names,
-                              gint                 n_variable_names)
+                              gint                 n_variable_names,
+                              gboolean             threadsafe)
 {
 	TrackerDBCursor *cursor;
+	TrackerDBInterface *iface;
+
+	/* As soon as we create a cursor, check if we need a collator reset
+	 * and notify the iface about the new cursor */
+	iface = ref_stmt->db_interface;
+	if (g_atomic_int_exchange_and_add (&iface->n_active_cursors, 1) == 0 &&
+	    g_atomic_int_compare_and_exchange (&(iface->collator_reset_requested), TRUE, FALSE)) {
+		tracker_db_interface_sqlite_reset_collator (iface);
+	}
 
 	cursor = g_object_new (TRACKER_TYPE_DB_CURSOR, NULL);
 
-	cursor->stmt = sqlite_stmt;
 	cursor->finished = FALSE;
 
-	if (ref_stmt) {
-		ref_stmt->stmt_is_sunk = TRUE;
-		cursor->ref_stmt = g_object_ref (ref_stmt);
-	} else {
-		cursor->ref_stmt = NULL;
-	}
+	cursor->threadsafe = threadsafe;
+
+	cursor->stmt = sqlite_stmt;
+	ref_stmt->stmt_is_sunk = TRUE;
+	cursor->ref_stmt = g_object_ref (ref_stmt);
 
 	if (types) {
 		gint i;
@@ -1465,7 +1431,6 @@ void
 tracker_db_cursor_rewind (TrackerDBCursor *cursor)
 {
 	g_return_if_fail (TRACKER_IS_DB_CURSOR (cursor));
-	g_return_if_fail (!cursor->ref_stmt->db_interface->pending);
 
 	sqlite3_reset (cursor->stmt);
 	cursor->finished = FALSE;
@@ -1476,8 +1441,6 @@ tracker_db_cursor_iter_next (TrackerDBCursor *cursor,
                              GCancellable    *cancellable,
                              GError         **error)
 {
-	g_return_val_if_fail (!cursor->ref_stmt->db_interface->pending, FALSE);
-
 	return db_cursor_iter_next (cursor, cancellable, error);
 }
 
@@ -1493,6 +1456,10 @@ db_cursor_iter_next (TrackerDBCursor *cursor,
 	if (!cursor->finished) {
 		guint result;
 
+		if (cursor->threadsafe) {
+			tracker_db_manager_lock ();
+		}
+
 		if (g_cancellable_is_cancelled (cancellable)) {
 			result = SQLITE_INTERRUPT;
 			sqlite3_reset (cursor->stmt);
@@ -1509,7 +1476,6 @@ db_cursor_iter_next (TrackerDBCursor *cursor,
 			             TRACKER_DB_INTERRUPTED,
 			             "Interrupted");
 		} else if (result != SQLITE_ROW && result != SQLITE_DONE) {
-
 			g_set_error (error,
 			             TRACKER_DB_INTERFACE_ERROR,
 			             TRACKER_DB_QUERY_ERROR,
@@ -1517,6 +1483,10 @@ db_cursor_iter_next (TrackerDBCursor *cursor,
 		}
 
 		cursor->finished = (result != SQLITE_ROW);
+
+		if (cursor->threadsafe) {
+			tracker_db_manager_unlock ();
+		}
 	}
 
 	return (!cursor->finished);
@@ -1637,13 +1607,13 @@ tracker_db_cursor_get_string (TrackerDBCursor *cursor,
 	}
 }
 
-TrackerDBResultSet *
+void
 tracker_db_statement_execute (TrackerDBStatement  *stmt,
                               GError             **error)
 {
-	g_return_val_if_fail (!stmt->stmt_is_sunk, NULL);
+	g_return_if_fail (!stmt->stmt_is_sunk);
 
-	return create_result_set_from_stmt (stmt->db_interface, stmt->stmt, NULL, error);
+	execute_stmt (stmt->db_interface, stmt->stmt, NULL, error);
 }
 
 TrackerDBCursor *
@@ -1652,7 +1622,7 @@ tracker_db_statement_start_cursor (TrackerDBStatement  *stmt,
 {
 	g_return_val_if_fail (!stmt->stmt_is_sunk, NULL);
 
-	return tracker_db_cursor_sqlite_new (stmt->stmt, stmt, NULL, 0, NULL, 0);
+	return tracker_db_cursor_sqlite_new (stmt->stmt, stmt, NULL, 0, NULL, 0, FALSE);
 }
 
 TrackerDBCursor *
@@ -1661,11 +1631,12 @@ tracker_db_statement_start_sparql_cursor (TrackerDBStatement   *stmt,
                                           gint                  n_types,
                                           const gchar         **variable_names,
                                           gint                  n_variable_names,
+                                          gboolean              threadsafe,
                                           GError              **error)
 {
 	g_return_val_if_fail (!stmt->stmt_is_sunk, NULL);
 
-	return tracker_db_cursor_sqlite_new (stmt->stmt, stmt, types, n_types, variable_names, n_variable_names);
+	return tracker_db_cursor_sqlite_new (stmt->stmt, stmt, types, n_types, variable_names, n_variable_names, threadsafe);
 }
 
 static void
