@@ -36,6 +36,7 @@
 #include <libtracker-common/tracker-date-time.h>
 #include <libtracker-common/tracker-file-utils.h>
 #include <libtracker-common/tracker-utils.h>
+#include <libtracker-common/tracker-locale.h>
 
 #if HAVE_TRACKER_FTS
 #include <libtracker-fts/tracker-fts.h>
@@ -56,7 +57,7 @@
 #define TRACKER_DB_PAGE_SIZE_DONT_SET -1
 
 /* Set current database version we are working with */
-#define TRACKER_DB_VERSION_NOW        TRACKER_DB_VERSION_0_9_24
+#define TRACKER_DB_VERSION_NOW        TRACKER_DB_VERSION_0_9_34
 #define TRACKER_DB_VERSION_FILE       "db-version.txt"
 #define TRACKER_DB_LOCALE_FILE        "db-locale.txt"
 
@@ -94,7 +95,8 @@ typedef enum {
 	TRACKER_DB_VERSION_0_9_16,  /* Fix for NB#184823 */
 	TRACKER_DB_VERSION_0_9_19,  /* collation */
 	TRACKER_DB_VERSION_0_9_21,  /* Fix for NB#186055 */
-	TRACKER_DB_VERSION_0_9_24   /* nmo:PhoneMessage class */
+	TRACKER_DB_VERSION_0_9_24,  /* nmo:PhoneMessage class */
+	TRACKER_DB_VERSION_0_9_34   /* ontology cache */
 } TrackerDBVersion;
 
 typedef struct {
@@ -158,6 +160,10 @@ static guint                 u_cache_size;
 
 static GStaticPrivate        interface_data_key = G_STATIC_PRIVATE_INIT;
 
+/* mutex used by singleton connection in libtracker-direct, not used by tracker-store */
+static GStaticMutex          global_mutex = G_STATIC_MUTEX_INIT;
+static TrackerDBInterface   *global_iface;
+
 static const gchar *
 location_to_directory (TrackerDBLocation location)
 {
@@ -178,17 +184,12 @@ db_exec_no_reply (TrackerDBInterface *iface,
                   const gchar        *query,
                   ...)
 {
-	TrackerDBResultSet *result_set;
 	va_list                     args;
 
 	va_start (args, query);
-	result_set = tracker_db_interface_execute_vquery (iface, NULL, query, args);
+	tracker_db_interface_execute_vquery (iface, NULL, query, args);
 	va_end (args);
 
-	if (result_set) {
-		g_object_unref (result_set);
-	}
-
 	return TRUE;
 }
 
@@ -209,30 +210,45 @@ db_set_params (TrackerDBInterface *iface,
                gint                cache_size,
                gint                page_size)
 {
-	TrackerDBResultSet *result_set;
+	gchar *queries = NULL;
+	const gchar *pragmas_file;
 
-	tracker_db_interface_execute_query (iface, NULL, "PRAGMA synchronous = OFF;");
-	tracker_db_interface_execute_query (iface, NULL, "PRAGMA count_changes = 0;");
-	tracker_db_interface_execute_query (iface, NULL, "PRAGMA temp_store = FILE;");
-	tracker_db_interface_execute_query (iface, NULL, "PRAGMA encoding = \"UTF-8\"");
-	tracker_db_interface_execute_query (iface, NULL, "PRAGMA auto_vacuum = 0;");
+	pragmas_file = g_getenv ("TRACKER_PRAGMAS_FILE");
 
-	result_set = tracker_db_interface_execute_query (iface, NULL, "PRAGMA journal_mode = WAL;");
-	if (result_set == NULL) {
-		/* Don't just silence the problem. This pragma must return 'WAL' */
-		g_message ("Can't set journal mode to WAL");
+	if (pragmas_file && g_file_get_contents (pragmas_file, &queries, NULL, NULL)) {
+		gchar *query = strtok (queries, "\n");
+		g_debug ("PRAGMA's from file: %s", pragmas_file);
+		while (query) {
+			g_debug ("  INIT query: %s", query);
+			tracker_db_interface_execute_query (iface, NULL, "%s", query);
+			query = strtok (NULL, "\n");
+		}
+		g_free (queries);
 	} else {
-		g_object_unref (result_set);
+		GError *error = NULL;
+
+		tracker_db_interface_execute_query (iface, NULL, "PRAGMA synchronous = OFF;");
+		tracker_db_interface_execute_query (iface, NULL, "PRAGMA count_changes = 0;");
+		tracker_db_interface_execute_query (iface, NULL, "PRAGMA temp_store = FILE;");
+		tracker_db_interface_execute_query (iface, NULL, "PRAGMA encoding = \"UTF-8\"");
+		tracker_db_interface_execute_query (iface, NULL, "PRAGMA auto_vacuum = 0;");
+
+		tracker_db_interface_execute_query (iface, &error, "PRAGMA journal_mode = WAL;");
+		if (error) {
+			/* Don't just silence the problem. This pragma must return 'WAL' */
+			g_message ("Can't set journal mode to WAL: '%s'",
+			           error->message);
+			g_clear_error (&error);
+		}
+
+		if (page_size != TRACKER_DB_PAGE_SIZE_DONT_SET) {
+			g_message ("  Setting page size to %d", page_size);
+			tracker_db_interface_execute_query (iface, NULL, "PRAGMA page_size = %d", page_size);
+		}
+
+		tracker_db_interface_execute_query (iface, NULL, "PRAGMA cache_size = %d", cache_size);
+		g_message ("  Setting cache size to %d", cache_size);
 	}
-
-	if (page_size != TRACKER_DB_PAGE_SIZE_DONT_SET) {
-		g_message ("  Setting page size to %d", page_size);
-		tracker_db_interface_execute_query (iface, NULL, "PRAGMA page_size = %d", page_size);
-	}
-
-	tracker_db_interface_execute_query (iface, NULL, "PRAGMA cache_size = %d", cache_size);
-	g_message ("  Setting cache size to %d", cache_size);
-
 }
 
 
@@ -566,7 +582,7 @@ tracker_db_manager_locale_changed (void)
 	gboolean changed;
 
 	/* Get current collation locale */
-	current_locale = setlocale (LC_COLLATE, NULL);
+	current_locale = tracker_locale_get (TRACKER_LOCALE_COLLATE);
 
 	/* Get db locale */
 	db_locale = db_get_locale ();
@@ -584,20 +600,20 @@ tracker_db_manager_locale_changed (void)
 	}
 
 	g_free (db_locale);
+	g_free (current_locale);
 	return changed;
 }
 
 void
 tracker_db_manager_set_current_locale (void)
 {
-	const gchar *current_locale;
+	gchar *current_locale;
 
 	/* Get current collation locale */
-	current_locale = setlocale (LC_COLLATE, NULL);
-
+	current_locale = tracker_locale_get (TRACKER_LOCALE_COLLATE);
 	g_message ("Changing db locale to: '%s'", current_locale);
-
 	db_set_locale (current_locale);
+	g_free (current_locale);
 }
 
 static void
@@ -641,6 +657,7 @@ static void
 db_recreate_all (void)
 {
 	guint i;
+	gchar *locale;
 
 	/* We call an internal version of this function here
 	 * because at the time 'initialized' = FALSE and that
@@ -666,8 +683,10 @@ db_recreate_all (void)
 		dbs[i].iface = NULL;
 	}
 
+	locale = tracker_locale_get (TRACKER_LOCALE_COLLATE);
 	/* Initialize locale file */
-	db_set_locale (setlocale (LC_COLLATE, NULL));
+	db_set_locale (locale);
+	g_free (locale);
 }
 
 void
@@ -705,7 +724,8 @@ tracker_db_manager_init (TrackerDBManagerFlags  flags,
                          guint                  select_cache_size,
                          guint                  update_cache_size,
                          TrackerBusyCallback    busy_callback,
-                         gpointer               busy_user_data)
+                         gpointer               busy_user_data,
+                         const gchar           *busy_operation)
 {
 	GType               etype;
 	TrackerDBVersion    version;
@@ -895,6 +915,7 @@ tracker_db_manager_init (TrackerDBManagerFlags  flags,
 				struct stat st;
 				GError *error = NULL;
 				TrackerDBStatement *stmt;
+				gchar *busy_status;
 
 				if (g_stat (dbs[i].abs_filename, &st) == 0) {
 					size = st.st_size;
@@ -913,10 +934,15 @@ tracker_db_manager_init (TrackerDBManagerFlags  flags,
 
 				loaded = TRUE;
 
+				/* Report OPERATION - STATUS */
+				busy_status = g_strdup_printf ("%s - %s",
+				                               busy_operation,
+				                               "Integrity checking");
 				tracker_db_interface_set_busy_handler (dbs[i].iface,
 				                                       busy_callback,
-				                                       "Integrity checking",
+				                                       busy_status,
 				                                       busy_user_data);
+				g_free (busy_status);
 
 				stmt = tracker_db_interface_create_statement (dbs[i].iface, TRACKER_DB_STATEMENT_CACHE_TYPE_NONE, &error,
 				                                              "PRAGMA integrity_check(1)");
@@ -1001,6 +1027,8 @@ tracker_db_manager_init (TrackerDBManagerFlags  flags,
 	if (flags & TRACKER_DB_MANAGER_READONLY) {
 		resources_iface = tracker_db_manager_get_db_interfaces_ro (1,
 		                                                           TRACKER_DB_METADATA);
+		/* libtracker-direct does not use per-thread interfaces */
+		global_iface = resources_iface;
 	} else {
 		resources_iface = tracker_db_manager_get_db_interfaces (1,
 		                                                        TRACKER_DB_METADATA);
@@ -1017,7 +1045,9 @@ tracker_db_manager_init (TrackerDBManagerFlags  flags,
 	s_cache_size = select_cache_size;
 	u_cache_size = update_cache_size;
 
-	g_static_private_set (&interface_data_key, resources_iface, (GDestroyNotify) g_object_unref);
+	if ((flags & TRACKER_DB_MANAGER_READONLY) == 0) {
+		g_static_private_set (&interface_data_key, resources_iface, (GDestroyNotify) g_object_unref);
+	}
 
 	return TRUE;
 }
@@ -1051,6 +1081,12 @@ tracker_db_manager_shutdown (void)
 	sys_tmp_dir = NULL;
 	g_free (sql_dir);
 
+	if (global_iface) {
+		/* libtracker-direct */
+		g_object_unref (global_iface);
+		global_iface = NULL;
+	}
+
 	/* shutdown db interfaces in all threads */
 	g_static_private_free (&interface_data_key);
 
@@ -1468,6 +1504,12 @@ tracker_db_manager_get_db_interface (void)
 	TrackerDBInterface *interface;
 
 	g_return_val_if_fail (initialized != FALSE, NULL);
+
+	if (global_iface) {
+		/* libtracker-direct */
+		return global_iface;
+	}
+
 	interface = g_static_private_get (&interface_data_key);
 
 	/* Ensure the interface is there */
@@ -1662,3 +1704,21 @@ tracker_db_manager_set_last_crawl_done (gboolean done)
 
 	g_free (filename);
 }
+
+void
+tracker_db_manager_lock (void)
+{
+	g_static_mutex_lock (&global_mutex);
+}
+
+gboolean
+tracker_db_manager_trylock (void)
+{
+	return g_static_mutex_trylock (&global_mutex);
+}
+
+void
+tracker_db_manager_unlock (void)
+{
+	g_static_mutex_unlock (&global_mutex);
+}
