@@ -45,8 +45,11 @@ struct _TrackerPropertyPrivate {
 	gchar         *name;
 	gchar         *table_name;
 
+	gboolean       use_gvdb;
+
 	TrackerPropertyType  data_type;
 	TrackerClass   *domain;
+	TrackerClass   *domain_index;
 	TrackerClass   *range;
 	gint           weight;
 	gint           id;
@@ -54,15 +57,18 @@ struct _TrackerPropertyPrivate {
 	TrackerProperty *secondary_index;
 	gboolean       fulltext_indexed;
 	gboolean       fulltext_no_limit;
-	gboolean       embedded;
 	gboolean       multiple_values;
+	gboolean       last_multiple_values;
 	gboolean       transient;
 	gboolean       is_inverse_functional_property;
 	gboolean       is_new;
 	gboolean       db_schema_changed;
 	gboolean       writeback;
+	gchar         *default_value;
+	GPtrArray     *is_new_domain_index;
 
 	GArray        *super_properties;
+	GArray        *domain_indexes;
 };
 
 static void property_finalize     (GObject      *object);
@@ -128,10 +134,10 @@ tracker_property_init (TrackerProperty *property)
 
 	priv->id = 0;
 	priv->weight = 1;
-	priv->embedded = TRUE;
 	priv->transient = FALSE;
 	priv->multiple_values = TRUE;
 	priv->super_properties = g_array_new (TRUE, TRUE, sizeof (TrackerProperty *));
+	priv->domain_indexes = g_array_new (TRUE, TRUE, sizeof (TrackerClass *));
 
 	/* Make GET_PRIV working */
 	property->priv = priv;
@@ -148,6 +154,10 @@ property_finalize (GObject *object)
 	g_free (priv->name);
 	g_free (priv->table_name);
 
+	if (priv->is_new_domain_index) {
+		g_ptr_array_unref (priv->is_new_domain_index);
+	}
+
 	if (priv->domain) {
 		g_object_unref (priv->domain);
 	}
@@ -161,6 +171,9 @@ property_finalize (GObject *object)
 	}
 
 	g_array_free (priv->super_properties, TRUE);
+	g_array_free (priv->domain_indexes, TRUE);
+
+	g_free (priv->default_value);
 
 	(G_OBJECT_CLASS (tracker_property_parent_class)->finalize) (object);
 }
@@ -173,12 +186,18 @@ property_finalize (GObject *object)
  * Returns: The newly created #TrackerProperty
  **/
 TrackerProperty *
-tracker_property_new (void)
+tracker_property_new (gboolean use_gvdb)
 {
 	TrackerProperty *property;
+	TrackerPropertyPrivate *priv;
 
 	property = g_object_new (TRACKER_TYPE_PROPERTY, NULL);
 
+	if (use_gvdb) {
+		priv = GET_PRIV (property);
+		priv->use_gvdb = use_gvdb;
+	}
+
 	return property;
 }
 
@@ -229,12 +248,12 @@ tracker_property_get_table_name (TrackerProperty *property)
 	priv = GET_PRIV (property);
 
 	if (!priv->table_name) {
-		if (priv->multiple_values) {
+		if (tracker_property_get_multiple_values (property)) {
 			priv->table_name = g_strdup_printf ("%s_%s",
-				tracker_class_get_name (priv->domain),
-				priv->name);
+				tracker_class_get_name (tracker_property_get_domain (property)),
+				tracker_property_get_name (property));
 		} else {
-			priv->table_name = g_strdup (tracker_class_get_name (priv->domain));
+			priv->table_name = g_strdup (tracker_class_get_name (tracker_property_get_domain (property)));
 		}
 	}
 
@@ -250,6 +269,27 @@ tracker_property_get_data_type (TrackerProperty *property)
 
 	priv = GET_PRIV (property);
 
+	if (priv->use_gvdb) {
+		const gchar *range_uri;
+
+		range_uri = tracker_ontologies_get_property_string_gvdb (priv->uri, "range");
+		if (strcmp (range_uri, XSD_STRING) == 0) {
+			priv->data_type = TRACKER_PROPERTY_TYPE_STRING;
+		} else if (strcmp (range_uri, XSD_BOOLEAN) == 0) {
+			priv->data_type = TRACKER_PROPERTY_TYPE_BOOLEAN;
+		} else if (strcmp (range_uri, XSD_INTEGER) == 0) {
+			priv->data_type = TRACKER_PROPERTY_TYPE_INTEGER;
+		} else if (strcmp (range_uri, XSD_DOUBLE) == 0) {
+			priv->data_type = TRACKER_PROPERTY_TYPE_DOUBLE;
+		} else if (strcmp (range_uri, XSD_DATE) == 0) {
+			priv->data_type = TRACKER_PROPERTY_TYPE_DATE;
+		} else if (strcmp (range_uri, XSD_DATETIME) == 0) {
+			priv->data_type = TRACKER_PROPERTY_TYPE_DATETIME;
+		} else {
+			priv->data_type = TRACKER_PROPERTY_TYPE_RESOURCE;
+		}
+	}
+
 	return priv->data_type;
 }
 
@@ -265,9 +305,53 @@ tracker_property_get_domain (TrackerProperty *property)
 
 	priv = GET_PRIV (property);
 
+	if (!priv->domain && priv->use_gvdb) {
+		const gchar *domain_uri;
+
+		domain_uri = tracker_ontologies_get_property_string_gvdb (priv->uri, "domain");
+		priv->domain = g_object_ref (tracker_ontologies_get_class_by_uri (domain_uri));
+	}
+
 	return priv->domain;
 }
 
+TrackerClass **
+tracker_property_get_domain_indexes (TrackerProperty *property)
+{
+	TrackerPropertyPrivate *priv;
+
+	/* Removed for performance:
+	 g_return_val_if_fail (TRACKER_IS_PROPERTY (property), NULL); */
+
+	g_return_val_if_fail (property != NULL, NULL);
+
+	priv = GET_PRIV (property);
+
+	if (priv->use_gvdb) {
+		TrackerClass *domain_index;
+		GVariant *variant;
+
+		tracker_property_reset_domain_indexes (property);
+
+		variant = tracker_ontologies_get_property_value_gvdb (priv->uri, "domain-indexes");
+		if (variant) {
+			GVariantIter iter;
+			const gchar *uri;
+
+			g_variant_iter_init (&iter, variant);
+			while (g_variant_iter_loop (&iter, "&s", &uri)) {
+				domain_index = tracker_ontologies_get_class_by_uri (uri);
+
+				tracker_property_add_domain_index (property, domain_index);
+			}
+
+			g_variant_unref (variant);
+		}
+	}
+
+	return (TrackerClass ** ) priv->domain_indexes->data;
+}
+
 TrackerClass *
 tracker_property_get_range (TrackerProperty *property)
 {
@@ -277,6 +361,13 @@ tracker_property_get_range (TrackerProperty *property)
 
 	priv = GET_PRIV (property);
 
+	if (!priv->range && priv->use_gvdb) {
+		const gchar *range_uri;
+
+		range_uri = tracker_ontologies_get_property_string_gvdb (priv->uri, "range");
+		priv->range = g_object_ref (tracker_ontologies_get_class_by_uri (range_uri));
+	}
+
 	return priv->range;
 }
 
@@ -367,6 +458,31 @@ tracker_property_get_is_new (TrackerProperty *property)
 	return priv->is_new;
 }
 
+gboolean
+tracker_property_get_is_new_domain_index (TrackerProperty *property,
+                                          TrackerClass    *class)
+{
+	TrackerPropertyPrivate *priv;
+	guint i;
+
+	g_return_val_if_fail (TRACKER_IS_PROPERTY (property), FALSE);
+	g_return_val_if_fail (TRACKER_IS_CLASS (class), FALSE);
+
+	priv = GET_PRIV (property);
+
+	if (!priv->is_new_domain_index) {
+		return FALSE;
+	}
+
+	for (i = 0; i < priv->is_new_domain_index->len; i++) {
+		if (g_ptr_array_index (priv->is_new_domain_index, i) == class) {
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
 gboolean
 tracker_property_get_writeback (TrackerProperty *property)
 {
@@ -391,18 +507,6 @@ tracker_property_get_db_schema_changed (TrackerProperty *property)
 	return priv->db_schema_changed;
 }
 
-gboolean
-tracker_property_get_embedded (TrackerProperty *property)
-{
-	TrackerPropertyPrivate *priv;
-
-	g_return_val_if_fail (TRACKER_IS_PROPERTY (property), FALSE);
-
-	priv = GET_PRIV (property);
-
-	return priv->embedded;
-}
-
 gboolean
 tracker_property_get_multiple_values (TrackerProperty *property)
 {
@@ -412,9 +516,36 @@ tracker_property_get_multiple_values (TrackerProperty *property)
 
 	priv = GET_PRIV (property);
 
+	if (priv->use_gvdb) {
+		GVariant *value;
+		gboolean result;
+
+		value = tracker_ontologies_get_property_value_gvdb (priv->uri, "max-cardinality");
+		if (value != NULL) {
+			result = FALSE;
+			g_variant_unref (value);
+		} else {
+			result = TRUE;
+		}
+
+		return result;
+	}
+
 	return priv->multiple_values;
 }
 
+gboolean
+tracker_property_get_last_multiple_values (TrackerProperty *property)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_val_if_fail (TRACKER_IS_PROPERTY (property), FALSE);
+
+	priv = GET_PRIV (property);
+
+	return priv->last_multiple_values;
+}
+
 gboolean
 tracker_property_get_is_inverse_functional_property (TrackerProperty *property)
 {
@@ -424,6 +555,21 @@ tracker_property_get_is_inverse_functional_property (TrackerProperty *property)
 
 	priv = GET_PRIV (property);
 
+	if (priv->use_gvdb) {
+		GVariant *value;
+		gboolean result;
+
+		value = tracker_ontologies_get_property_value_gvdb (priv->uri, "inverse-functional");
+		if (value != NULL) {
+			result = g_variant_get_boolean (value);
+			g_variant_unref (value);
+		} else {
+			result = FALSE;
+		}
+
+		return result;
+	}
+
 	return priv->is_inverse_functional_property;
 }
 
@@ -439,6 +585,18 @@ tracker_property_get_super_properties (TrackerProperty *property)
 	return (TrackerProperty **) priv->super_properties->data;
 }
 
+const gchar *
+tracker_property_get_default_value (TrackerProperty *property)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_val_if_fail (TRACKER_IS_PROPERTY (property), NULL);
+
+	priv = GET_PRIV (property);
+
+	return priv->default_value;
+}
+
 void
 tracker_property_set_uri (TrackerProperty *property,
                           const gchar     *value)
@@ -491,7 +649,6 @@ tracker_property_set_transient (TrackerProperty *property,
 	priv = GET_PRIV (property);
 
 	priv->transient = value;
-	priv->multiple_values = TRUE;
 }
 
 void
@@ -514,6 +671,60 @@ tracker_property_set_domain (TrackerProperty *property,
 	}
 }
 
+void
+tracker_property_add_domain_index (TrackerProperty *property,
+                                   TrackerClass    *value)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_if_fail (TRACKER_IS_PROPERTY (property));
+	g_return_if_fail (TRACKER_IS_CLASS (value));
+
+	priv = GET_PRIV (property);
+
+	g_array_append_val (priv->domain_indexes, value);
+}
+
+void
+tracker_property_del_domain_index (TrackerProperty *property,
+                                   TrackerClass    *value)
+{
+	TrackerClass **classes;
+	TrackerPropertyPrivate *priv;
+	gint i = 0, found = -1;
+
+	g_return_if_fail (TRACKER_IS_PROPERTY (property));
+	g_return_if_fail (TRACKER_IS_CLASS (value));
+
+	priv = GET_PRIV (property);
+
+	classes = (TrackerClass **) priv->domain_indexes->data;
+	while (*classes) {
+		if (*classes == value) {
+			found = i;
+			break;
+		}
+		i++;
+		classes++;
+	}
+
+	if (found != -1) {
+		g_array_remove_index (priv->domain_indexes, found);
+	}
+}
+
+void
+tracker_property_reset_domain_indexes (TrackerProperty *property)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_if_fail (TRACKER_IS_PROPERTY (property));
+
+	priv = GET_PRIV (property);
+	g_array_free (priv->domain_indexes, TRUE);
+	priv->domain_indexes = g_array_new (TRUE, TRUE, sizeof (TrackerClass *));
+}
+
 void
 tracker_property_set_secondary_index (TrackerProperty *property,
                                       TrackerProperty *value)
@@ -621,6 +832,53 @@ tracker_property_set_is_new (TrackerProperty *property,
 	priv->is_new = value;
 }
 
+void
+tracker_property_set_is_new_domain_index (TrackerProperty *property,
+                                          TrackerClass    *class,
+                                          gboolean         value)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_if_fail (TRACKER_IS_PROPERTY (property));
+
+	if (class) {
+		g_return_if_fail (TRACKER_IS_CLASS (class));
+	}
+
+	priv = GET_PRIV (property);
+
+	if (value) {
+		if (!priv->is_new_domain_index) {
+			priv->is_new_domain_index = g_ptr_array_new_with_free_func ((GDestroyNotify) g_object_unref);
+		}
+		g_ptr_array_add (priv->is_new_domain_index, g_object_ref (class));
+	} else {
+		guint i;
+		gboolean found = FALSE;
+
+		if (!priv->is_new_domain_index) {
+			return;
+		}
+
+		if (!class) {
+			g_ptr_array_unref (priv->is_new_domain_index);
+			priv->is_new_domain_index = NULL;
+			return;
+		}
+
+		for (i = 0; i < priv->is_new_domain_index->len; i++) {
+			if (g_ptr_array_index (priv->is_new_domain_index, i) == class) {
+				found = TRUE;
+				break;
+			}
+		}
+
+		if (found) {
+			g_ptr_array_remove_index (priv->is_new_domain_index, i);
+		}
+	}
+}
+
 void
 tracker_property_set_writeback (TrackerProperty *property,
                                 gboolean         value)
@@ -673,19 +931,6 @@ tracker_property_set_fulltext_no_limit (TrackerProperty *property,
 	priv->fulltext_no_limit = value;
 }
 
-void
-tracker_property_set_embedded (TrackerProperty *property,
-                               gboolean         value)
-{
-	TrackerPropertyPrivate *priv;
-
-	g_return_if_fail (TRACKER_IS_PROPERTY (property));
-
-	priv = GET_PRIV (property);
-
-	priv->embedded = value;
-}
-
 void
 tracker_property_set_multiple_values (TrackerProperty *property,
                                       gboolean         value)
@@ -696,11 +941,20 @@ tracker_property_set_multiple_values (TrackerProperty *property,
 
 	priv = GET_PRIV (property);
 
-	if (priv->transient) {
-		priv->multiple_values = TRUE;
-	} else {
-		priv->multiple_values = value;
-	}
+	priv->multiple_values = value;
+}
+
+void
+tracker_property_set_last_multiple_values (TrackerProperty *property,
+                                           gboolean         value)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_if_fail (TRACKER_IS_PROPERTY (property));
+
+	priv = GET_PRIV (property);
+
+	priv->last_multiple_values = value;
 }
 
 void
@@ -730,3 +984,16 @@ tracker_property_add_super_property (TrackerProperty *property,
 	g_array_append_val (priv->super_properties, value);
 }
 
+void
+tracker_property_set_default_value (TrackerProperty *property,
+                                    const gchar     *value)
+{
+	TrackerPropertyPrivate *priv;
+
+	g_return_if_fail (TRACKER_IS_PROPERTY (property));
+
+	priv = GET_PRIV (property);
+
+	g_free (priv->default_value);
+	priv->default_value = g_strdup (value);
+}
